from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.unary_operator import UnaryOperator


class NotOperator(UnaryOperator):
    def unary(self, evaluator: Evaluator, arg: object):
        evaluator.boolean_convert(arg)
        return evaluator.boolean_convert(arg) is not True



import base64
import collections
import hashlib
import threading
from concurrent.futures import Future
from typing import Optional

from sdk.audience_matcher import AudienceMatcher
from sdk.context_config import ContextConfig
from sdk.context_data_provider import ContextDataProvider
from sdk.context_event_handler import ContextEventHandler
from sdk.context_event_logger import ContextEventLogger, EventType
from sdk.internal.lock.atomic_bool import AtomicBool
from sdk.internal.lock.atomic_int import AtomicInt
from sdk.internal.lock.concurrency import Concurrency
from sdk.internal.lock.read_write_lock import ReadWriteLock
from sdk.internal.variant_assigner import VariantAssigner
from sdk.json.attribute import Attribute
from sdk.json.context_data import ContextData
from sdk.json.experiment import Experiment
from sdk.json.exposure import Exposure
from sdk.json.goal_achievement import GoalAchievement
from sdk.json.publish_event import PublishEvent
from sdk.json.unit import Unit
from sdk.time.clock import Clock
from sdk.variable_parser import VariableParser


class Assignment:
    def __init__(self):
        self.id: Optional[int] = 0
        self.iteration: Optional[int] = 0
        self.full_on_variant: Optional[int] = 0
        self.name: Optional[str] = None
        self.unit_type: Optional[str] = None
        self.traffic_split: list[int] = []
        self.variant: Optional[int] = 0
        self.assigned: Optional[bool] = False
        self.overridden: Optional[bool] = False
        self.eligible: Optional[bool] = False
        self.full_on: Optional[bool] = False
        self.custom: Optional[bool] = False
        self.audience_mismatch: Optional[bool] = False
        self.variables: dict = {}
        self.exposed = AtomicBool()


class ExperimentVariables:
    data: Optional[Experiment]
    variables: Optional[list[dict]]

class ContextCustomFieldValue:
    type: Optional[str]
    value: Optional[any]

def experiment_matches(experiment: Experiment, assignment: Assignment):
    return experiment.id == assignment.id and \
           experiment.unitType == assignment.unit_type and \
           experiment.iteration == assignment.iteration and \
           experiment.fullOnVariant == assignment.full_on_variant and \
           collections.Counter(experiment.trafficSplit) == \
           collections.Counter(assignment.traffic_split)


class Context:
    def __init__(self,
                 clock: Clock, config: ContextConfig,
                 data_future: Future, data_provider: ContextDataProvider,
                 event_handler: ContextEventHandler,
                 event_logger: ContextEventLogger,
                 variable_parser: VariableParser,
                 audience_matcher: AudienceMatcher):
        self.clock = clock
        self.publish_delay = config.publish_delay
        self.refresh_interval = config.refresh_interval
        self.event_handler = event_handler
        self.event_logger = event_logger
        self.data_provider = data_provider
        self.variable_parser = variable_parser
        self.audience_matcher = audience_matcher

        self.units = {}
        self.index = {}
        self.index_variables = {}
        self.context_custom_fields = {}
        self.assignment_cache = {}
        self.cassignments = {}
        self.overrides = {}

        self.exposures = []
        self.achievements = []

        self.data: Optional[ContextData] = None

        self.failed = False

        self.closed = AtomicBool()
        self.closing = AtomicBool()
        self.refreshing = AtomicBool()
        self.pending_count = AtomicInt()

        self.context_lock = ReadWriteLock()
        self.data_lock = ReadWriteLock()
        self.timeout_lock = ReadWriteLock()
        self.event_lock = ReadWriteLock()

        self.refresh_future: Optional[Future] = None
        self.closing_future: Optional[Future] = None

        self.refresh_timer: Optional[threading.Timer] = None
        self.timeout: Optional[threading.Timer] = None

        if config.units is not None:
            self.set_units(config.units)

        self.assigners = dict.fromkeys((range(len(self.units))))
        self.hashed_units = dict.fromkeys((range(len(self.units))))

        self.attributes: list[Attribute] = []

        if config.attributes is not None:
            self.set_attributes(config.attributes)

        if config.overrides is not None:
            self.overrides = dict(config.overrides)
        else:
            self.overrides = {}

        if config.cassigmnents is not None:
            self.cassignments = dict(config.cassigmnents)
        else:
            self.cassignments = {}

        if data_future.done():
            def when_finished(data: Future):
                if data.done() and data.cancelled() is False and \
                        data.exception() is None:
                    self.set_data(data.result())
                    self.log_event(EventType.READY, data.result())
                elif data.cancelled() is False and \
                        data.exception() is not None:
                    self.set_data_failed(data.exception())
                    self.log_error(data.exception())

            data_future.add_done_callback(when_finished)
        else:
            self.ready_future = Future()

            def when_finished(data: Future):
                if data.done() and data.cancelled() is False and \
                        data.exception() is None:
                    self.set_data(data.result())
                    self.ready_future.set_result(None)
                    self.ready_future = None
                    self.log_event(EventType.READY, data.result())

                    if self.get_pending_count() > 0:
                        self.set_timeout()
                elif data.cancelled() is False and \
                        data.exception() is not None:
                    self.set_data_failed(data.exception())
                    self.ready_future.set_result(None)
                    self.ready_future = None
                    self.log_error(data.exception())

            data_future.add_done_callback(when_finished)

    def set_units(self, units: dict):
        for key, value in units.items():
            self.set_unit(key, value)

    def set_unit(self, unit_type: str, uid: str):
        self.check_not_closed()

        try:
            self.context_lock.acquire_write()

            if unit_type in self.units.keys() and self.units[unit_type] != uid:
                raise ValueError("Unit already set.")

            trimmed = uid.strip()
            if len(trimmed) == 0:
                raise ValueError("Unit UID must not be blank.")

            self.units[unit_type] = trimmed
        finally:
            self.context_lock.release_write()

    def set_attributes(self, attributes: dict):
        for key, value in attributes.items():
            self.set_attribute(key, value)

    def set_attribute(self, name: str, value: object):
        self.check_not_closed()
        attribute = Attribute()
        attribute.name = name
        attribute.value = value
        attribute.setAt = self.clock.millis()
        Concurrency.add_rw(self.context_lock, self.attributes, attribute)

    def check_not_closed(self):
        if self.closed.value:
            raise RuntimeError('ABSmartly Context is closed')
        elif self.closing.value:
            raise RuntimeError('ABSmartly Context is closing')

    def set_data(self, data: ContextData):
        index = {}
        index_variables = {}
        context_custom_fields = {}

        for experiment in data.experiments:
            experiment_variables = ExperimentVariables()
            experiment_variables.data = experiment
            experiment_variables.variables = []

            for variant in experiment.variants:
                if variant.config is not None and len(variant.config) > 0:
                    variables = self.variable_parser.parse(
                        self,
                        experiment.name,
                        variant.name,
                        variant.config)
                    for key, value in variables.items():
                        index_variables[key] = experiment_variables
                    experiment_variables.variables.append(variables)
                else:
                    experiment_variables.variables.append({})
            index[experiment.name] = experiment_variables

            if experiment.customFieldValues is not None:
                experimentCustomFields = {}
                for customFieldValue in experiment.customFieldValues:
                    value = ContextCustomFieldValue()
                    value.type = customFieldValue.type

                    if customFieldValue.value is not None:
                        customValue = customFieldValue.value

                        if customFieldValue.type.startswith("json"):
                            value.value = self.variable_parser.parse(
                                self,
                                experiment.name,
                                customFieldValue.name,
                                customValue)

                        elif customFieldValue.type.startswith("boolean"):
                            value.value = bool(customValue)

                        elif customFieldValue.type.startswith("number"):
                            value.value = int(customValue)


                        else:
                            value.value = customValue

                    experimentCustomFields[customFieldValue.name] = value

                context_custom_fields[experiment.name] = experimentCustomFields

        try:
            self.data_lock.acquire_write()

            self.index = index
            self.index_variables = index_variables
            self.context_custom_fields = context_custom_fields
            self.data = data

            self.set_refresh_timer()
        finally:
            self.data_lock.release_write()

    def set_refresh_timer(self):
        if self.refresh_interval > 0 and self.refresh_timer is None and not self.is_closing() and not self.is_closed():
            def ref():
                self.refresh_async()
                self.refresh_timer = threading.Timer(
                    self.refresh_interval,
                    ref)
                self.refresh_timer.start()

            self.refresh_timer = threading.Timer(
                self.refresh_interval,
                ref)
            self.refresh_timer.start()

    def set_timeout(self):
        if self.is_ready():
            if self.timeout is None:
                try:
                    self.timeout_lock.acquire_write()

                    def flush():
                        self.flush()

                    self.timeout = threading.Timer(self.publish_delay, flush)
                    self.timeout.start()
                finally:
                    self.timeout_lock.release_write()

    def is_ready(self):
        return self.data is not None

    def is_failed(self):
        return self.failed

    def is_closed(self):
        return self.closed.value

    def is_closing(self):
        return not self.closed.value and self.closing.value

    def refresh_async(self):
        self.check_not_closed()

        if self.refreshing.compare_and_set(False, True):
            self.refresh_future = Future()

            def when_ready(data):
                if data.done() and data.cancelled() is False and \
                        data.exception() is None:
                    self.set_data(data.result())
                    self.refreshing.set(False)
                    self.refresh_future.set_result(None)
                    self.log_event(EventType.REFRESH, data.result())
                elif data.cancelled() is False and \
                        data.exception() is not None:
                    self.refreshing.set(False)
                    self.refresh_future.set_exception(data.exception())
                    self.log_error(data.exception())

            self.data_provider\
                .get_context_data()\
                .add_done_callback(when_ready)

        if self.refresh_future is not None:
            return self.refresh_future
        else:
            result = Future()
            result.set_result(None)
            return result

    def set_data_failed(self, exception):
        try:
            self.data_lock.acquire_write()
            self.index = {}
            self.index_variables = {}
            self.data = ContextData()
            self.failed = True
        finally:
            self.data_lock.release_write()

    def log_error(self, exception):
        if self.event_logger is not None:
            self.event_logger.handle_event(EventType.ERROR, exception)

    def log_event(self, event: EventType, data: object):
        if self.event_logger is not None:
            self.event_logger.handle_event(event, data)

    def get_pending_count(self):
        return self.pending_count.get()

    def flush(self):
        self.clear_timeout()

        if self.failed is False:
            if self.pending_count.get() > 0:
                exposures = None
                achievements = None
                event_count = 0
                try:
                    self.event_lock.acquire_write()
                    event_count = self.pending_count.get()

                    if event_count > 0:
                        if len(self.exposures) > 0:
                            exposures = list(self.exposures)
                            self.exposures.clear()

                        if len(self.achievements) > 0:
                            achievements = list(self.achievements)
                            self.achievements.clear()
                        self.pending_count.set(0)
                finally:
                    self.event_lock.release_write()

                if event_count > 0:
                    event = PublishEvent()
                    event.hashed = True
                    event.publishedAt = self.clock.millis()
                    event.units = []
                    for key, value in self.units.items():
                        unit = Unit()
                        unit.type = key
                        unit.uid = str(
                            self.get_unit_hash(key, value),
                            encoding='ascii')\
                            .encode('ascii', errors='ignore')\
                            .decode()
                        event.units.append(unit)
                    if len(self.attributes) > 0:
                        event.attributes = list(self.attributes)
                    else:
                        event.attributes = None
                    event.exposures = exposures
                    event.goals = achievements

                    result = Future()

                    def run(data):
                        if data.done() and \
                                data.cancelled() is False and \
                                data.exception() is None:
                            self.log_event(EventType.PUBLISH, event)
                            result.set_result(None)
                        elif data.cancelled() is False and \
                                data.exception() is not None:
                            self.log_error(data.exception())
                            result.set_exception(data.exception())

                    self.event_handler\
                        .publish(self, event)\
                        .add_done_callback(run)
                    return result
        else:
            try:
                self.event_lock.acquire_write()
                self.exposures.clear()
                self.achievements.clear()
                self.pending_count.set(0)
            finally:
                self.event_lock.release_write()

        result = Future()
        result.set_result(None)
        return result

    def close(self):
        self.close_async().result()

    def refresh(self):
        self.refresh_async().result()

    def publish(self):
        self.publish_async().result()

    def publish_async(self):
        self.check_not_closed()
        return self.flush()

    def track(self, goal_name: str, properties: dict):
        self.check_not_closed()

        achievement = GoalAchievement()
        achievement.achievedAt = self.clock.millis()
        achievement.name = goal_name
        if properties is None:
            achievement.properties = None
        else:
            achievement.properties = dict(properties)

        try:
            self.event_lock.acquire_write()
            self.pending_count.increment_and_get()
            self.achievements.append(achievement)
        finally:
            self.event_lock.release_write()

        self.log_event(EventType.GOAL, achievement)
        self.set_timeout()

    def wait_until_ready(self):
        if self.data is None:
            if self.ready_future is not None and not self.ready_future.done():
                self.ready_future.result()
        return self

    def wait_until_ready_async(self):
        if self.data is not None:
            result = Future()
            result.set_result(self)
            return result
        else:
            def apply(fut: Future):
                return self

            self.ready_future.add_done_callback(apply)
            return self.ready_future

    def clear_timeout(self):
        if self.timeout is not None:
            try:
                self.timeout_lock.acquire_write()
                if self.timeout is not None:
                    self.timeout.cancel()
                    self.timeout = None
            finally:
                self.timeout_lock.release_write()

    def clear_refresh_timer(self):
        if self.refresh_timer is not None:
            self.refresh_timer.cancel()
            self.refresh_timer = None

    def get_variable_value(self, key: str, default_value: object):
        self.check_ready(True)

        assignment = self.get_variable_assignment(key)
        if assignment is not None:
            if assignment.variables is not None:
                if not assignment.exposed.value:
                    self.queue_exposure(assignment)

                if key in assignment.variables:
                    return assignment.variables[key]
        return default_value

    def peek_variable_value(self, key: str, default_value: object):
        self.check_ready(True)

        assignment = self.get_variable_assignment(key)
        if assignment is not None:
            if assignment.variables is not None:
                if key in assignment.variables:
                    return assignment.variables[key]
        return default_value

    def peek_treatment(self, experiment_name: str):
        self.check_ready(True)

        return self.get_assignment(experiment_name).variant

    def get_unit_hash(self, unit_type: str, unit_uid: str):
        def computer(key: str):
            dig = hashlib.md5(unit_uid.encode('utf-8')).digest()
            unithash = base64.urlsafe_b64encode(dig).rstrip(b'=')
            return unithash

        return Concurrency.compute_if_absent_rw(
            self.context_lock,
            self.hashed_units,
            unit_type,
            computer)

    def get_treatment(self, experiment_name: str):
        self.check_ready(True)
        assignment = self.get_assignment(experiment_name)
        if not assignment.exposed.value:
            self.queue_exposure(assignment)
        return assignment.variant

    def get_variable_keys(self):
        self.check_ready(True)

        variable_keys = {}
        try:
            self.data_lock.acquire_read()
            for key, value in self.index_variables.items():
                expr_var: ExperimentVariables = value
                variable_keys[key] = expr_var.data.name
        finally:
            self.data_lock.release_write()

        return variable_keys

    def get_custom_field_keys(self):
        self.check_ready(True)

        keys = []
        try:
            self.data_lock.acquire_read()

            for experiment in self.data.experiments:
                customFieldValues = experiment.customFieldValues

                if customFieldValues is not None:
                    for customFieldValue in customFieldValues:
                        keys.append(customFieldValue.name)
        finally:
            self.data_lock.release_write()

        keys = list(set(keys))
        keys.sort()

        return keys

    def get_custom_field_value(self, experiment_name: str, key: str):
        self.check_ready(True)

        value: any = None
        try:
            self.data_lock.acquire_read()

            if experiment_name in self.context_custom_fields:
                custom_field_value = self.context_custom_fields[experiment_name]
                if key in custom_field_value:
                    value = custom_field_value[key].value

        finally:
            self.data_lock.release_read()

        return value

    def get_custom_field_type(self, experiment_name: str, key: str):
        self.check_ready(True)

        type = None
        try:
            self.data_lock.acquire_read()

            if experiment_name in self.context_custom_fields:
                customFieldValue = self.context_custom_fields[experiment_name]
                if key in customFieldValue:
                    type = customFieldValue[key].type

        finally:
            self.data_lock.release_read()

        return type

    def get_assignment(self, experiment_name: str):
        try:
            self.context_lock.acquire_read()

            if experiment_name in self.assignment_cache:
                assignment: Assignment = self.assignment_cache[experiment_name]

                experiment: ExperimentVariables = \
                    self.get_experiment(experiment_name)

                if experiment_name in self.overrides:
                    override = self.overrides[experiment_name]
                    if assignment.overridden and \
                            assignment.variant == override:
                        return assignment
                elif experiment is None:
                    if assignment.assigned is False:
                        return assignment
                elif experiment_name not in self.cassignments or \
                        self.cassignments[experiment_name] == \
                        assignment.variant:
                    if experiment_matches(experiment.data, assignment):
                        return assignment
        finally:
            self.context_lock.release_read()

        try:
            self.context_lock.acquire_write()
            experiment: ExperimentVariables = \
                self.get_experiment(experiment_name)
            assignment = Assignment()
            assignment.name = experiment_name
            assignment.eligible = True

            if experiment_name in self.overrides:
                if experiment is not None:
                    assignment.id = experiment.data.id
                    assignment.unit_type = experiment.data.unitType

                assignment.overridden = True
                assignment.variant = self.overrides[experiment_name]
            else:
                if experiment is not None:
                    unit_type = experiment.data.unitType

                    if experiment.data.audience is not None and \
                            len(experiment.data.audience) > 0:
                        attrs = {}
                        for attr in self.attributes:
                            attrs[attr.name] = attr.value
                        match = self.audience_matcher.evaluate(
                            experiment.data.audience,
                            attrs)
                        if match is not None:
                            assignment.audience_mismatch = not match.result
                    if experiment.data.audienceStrict and \
                            assignment.audience_mismatch:
                        assignment.variant = 0
                    elif experiment.data.fullOnVariant == 0:
                        if experiment.data.unitType in self.units:
                            uid = self.units[experiment.data.unitType]
                            unit_hash = self.get_unit_hash(unit_type, uid)
                            assigner: VariantAssigner = \
                                self.get_variant_assigner(unit_type,
                                                          unit_hash)
                            eligible = \
                                assigner.assign(
                                    experiment.data.trafficSplit,
                                    experiment.data.trafficSeedHi,
                                    experiment.data.trafficSeedLo) == 1
                            if eligible:
                                if experiment_name in self.cassignments:
                                    custom = self.cassignments[experiment_name]
                                    assignment.variant = custom
                                    assignment.custom = True
                                else:
                                    assignment.variant = \
                                        assigner.assign(experiment.data.split,
                                                        experiment.data.seedHi,
                                                        experiment.data.seedLo)
                            else:
                                assignment.eligible = False
                                assignment.variant = 0

                            assignment.assigned = True

                    else:
                        assignment.assigned = True
                        assignment.variant = experiment.data.fullOnVariant
                        assignment.full_on = True

                    assignment.unit_type = unit_type
                    assignment.id = experiment.data.id
                    assignment.iteration = experiment.data.iteration
                    assignment.traffic_split = experiment.data.trafficSplit
                    assignment.full_on_variant = experiment.data.fullOnVariant

            if experiment is not None and \
                    (assignment.variant < len(experiment.data.variants)):
                assignment.variables = experiment.variables[assignment.variant]

            self.assignment_cache[experiment_name] = assignment
            return assignment
        finally:
            self.context_lock.release_write()

    def check_ready(self, expect_not_closed: bool):
        if not self.is_ready():
            raise RuntimeError('ABSmartly Context is not yet ready')
        elif expect_not_closed:
            self.check_not_closed()

    def get_experiment(self, experiment_name: str):
        try:
            self.data_lock.acquire_read()
            return self.index.get(experiment_name, None)
        finally:
            self.data_lock.release_read()

    def get_experiments(self):
        self.check_ready(True)

        try:
            self.data_lock.acquire_read()
            experiment_names = []
            for experiment in self.data.experiments:
                experiment_names.append(experiment.name)

            return experiment_names
        finally:
            self.data_lock.release_read()

    def get_data(self):
        self.check_ready(True)

        try:
            self.data_lock.acquire_read()
            return self.data
        finally:
            self.data_lock.release_read()

    def set_override(self, experiment_name: str, variant: int):
        self.check_not_closed()

        return Concurrency.put_rw(self.context_lock,
                                  self.overrides,
                                  experiment_name, variant)

    def get_override(self, experiment_name: str):
        return Concurrency.get_rw(self.context_lock,
                                  self.overrides,
                                  experiment_name)

    def set_overrides(self, overrides: dict):
        for key, value in overrides.items():
            self.set_override(key, value)

    def set_custom_assignment(self, experiment_name: str, variant: int):
        self.check_not_closed()

        Concurrency.put_rw(self.context_lock,
                           self.cassignments,
                           experiment_name, variant)

    def get_custom_assignment(self, experiment_name: str):
        return Concurrency.get_rw(self.context_lock,
                                  self.cassignments,
                                  experiment_name)

    def set_custom_assignments(self, custom_assignments: dict):
        for key, value in custom_assignments.items():
            self.set_custom_assignment(key, value)

    def get_variant_assigner(self, unit_type: str, unit_hash: bytes):
        def apply(key: str):
            return VariantAssigner(bytearray(unit_hash))

        return Concurrency.compute_if_absent_rw(self.context_lock,
                                                self.assigners,
                                                unit_type, apply)

    def get_variable_experiment(self, key: str):
        return Concurrency.get_rw(self.data_lock, self.index_variables, key)

    def get_variable_assignment(self, key: str):
        experiment: ExperimentVariables = self.get_variable_experiment(key)
        if experiment is not None:
            return self.get_assignment(experiment.data.name)
        return None

    def close_async(self):
        if not self.closed.value:
            if self.closing.compare_and_set(False, True):
                self.clear_refresh_timer()

                if self.pending_count.get() > 0:
                    self.closing_future = Future()

                    def accept(res: Future):
                        if res.done() and res.cancelled() is False \
                                and res.exception() is None:
                            self.closed.set(True)
                            self.closing.set(False)
                            self.closing_future.set_result(None)
                            self.log_event(EventType.CLOSE, None)
                        elif res.cancelled() is False \
                                and res.exception() is not None:
                            self.closed.set(True)
                            self.closing.set(False)
                            self.closing_future.exception(res.exception())

                    self.flush().add_done_callback(accept)
                    return self.closing_future

                else:
                    self.closed.set(True)
                    self.closing.set(False)
                    self.log_event(EventType.CLOSE, None)

            if self.closing_future is not None:
                return self.closing_future

        result = Future()
        result.set_result(None)
        return result

    def queue_exposure(self, assignment: Assignment):
        if assignment.exposed.compare_and_set(False, True):
            exposure = Exposure()
            exposure.id = assignment.id
            exposure.name = assignment.name
            exposure.unit = assignment.unit_type
            exposure.variant = assignment.variant
            exposure.exposedAt = self.clock.millis()
            exposure.assigned = assignment.assigned
            exposure.eligible = assignment.eligible
            exposure.overridden = assignment.overridden
            exposure.fullOn = assignment.full_on
            exposure.custom = assignment.custom
            exposure.audienceMismatch = assignment.audience_mismatch

            try:
                self.event_lock.acquire_write()
                self.pending_count.increment_and_get()
                self.exposures.append(exposure)
            finally:
                self.event_lock.release_write()

            self.log_event(EventType.EXPOSURE, exposure)
            self.set_timeout()



from abc import abstractmethod

from sdk.json.context_data import ContextData


class ContextDataDeserializer:

    @abstractmethod
    def deserialize(self,
                    bytes_: bytearray,
                    offset: int,
                    length: int) -> ContextData:
        raise NotImplementedError



from typing import Optional

import jsons
from jsons import DeserializationError

from sdk.audience_deserializer import AudienceDeserializer


class DefaultAudienceDeserializer(AudienceDeserializer):
    def deserialize(self,
                    bytes_: bytes,
                    offset: int,
                    length: int) -> Optional[dict]:
        try:
            return jsons.loadb(bytes_, dict)
        except DeserializationError:
            return None



import typing

from sdk.json.custom_field_value import CustomFieldValue
from sdk.json.experiement_application import ExperimentApplication
from sdk.json.experiment_variant import ExperimentVariant


class Experiment:
    id: typing.Optional[int] = 0
    name: typing.Optional[str] = None
    unitType: typing.Optional[str] = None
    iteration: typing.Optional[int] = 0
    seedHi: typing.Optional[int] = 0
    seedLo: typing.Optional[int] = 0
    split: typing.Optional[list[float]] = None
    trafficSeedHi: typing.Optional[int] = 0
    trafficSeedLo: typing.Optional[int] = 0
    trafficSplit: typing.Optional[list[float]] = None
    fullOnVariant: typing.Optional[int] = 0
    applications: typing.Optional[list[ExperimentApplication]] = None
    variants: typing.Optional[list[ExperimentVariant]] = None
    audienceStrict: typing.Optional[bool] = False
    audience: typing.Optional[str] = None
    customFieldValues: typing.Optional[list[CustomFieldValue]] = None



from typing import Optional

from sdk.audience_deserializer import AudienceDeserializer
from sdk.client import Client
from sdk.context_data_provider import ContextDataProvider
from sdk.context_event_handler import ContextEventHandler
from sdk.context_event_logger import ContextEventLogger
from sdk.variable_parser import VariableParser


class ABSmartlyConfig:
    context_data_provider: Optional[ContextDataProvider] = None
    context_event_handler: Optional[ContextEventHandler] = None
    context_event_logger: Optional[ContextEventLogger] = None
    audience_deserializer: Optional[AudienceDeserializer] = None
    client: Optional[Client] = None
    variable_parser: Optional[VariableParser] = None



import unittest

from sdk.default_audience_deserializer import DefaultAudienceDeserializer


class DefaultAudienceDataDeserializerTest(unittest.TestCase):
    def test_deserialize(self):
        deser = DefaultAudienceDeserializer()
        audience = \
            "{\"filter\":[{\"gte\":[{\"var\":\"age\"},{\"value\":20.0}]}]}"
        actual = deser.deserialize(bytes(audience, encoding="utf-8"),
                                   0,
                                   len(audience))
        expected = {"filter": [{
            "gte": [{
                "var": "age"
            }, {
                "value": 20.0
            }]
        }]}
        self.assertEqual(expected, actual)

    def test_deserializer_incorrect(self):
        deser = DefaultAudienceDeserializer()
        audience = \
            "{\"filter\":[{\"gte\":[{\"var\":\"age\"},{\"value\":20.0}]]}"
        actual = deser.deserialize(bytes(audience, encoding="utf-8"),
                                   0,
                                   len(audience))
        self.assertEqual(None, actual)



from abc import abstractmethod
from concurrent.futures import Future
from typing import Optional

from sdk.json.context_data import ContextData
from sdk.json.publish_event import PublishEvent


class ContextEventHandler:

    @abstractmethod
    def publish(self, context, event: PublishEvent) -> \
            Future[Optional[ContextData]]:
        raise NotImplementedError



import threading

from sdk.internal import murmur32, buffers


class VariantAssigner:

    def __init__(self, unithash: bytearray):

        self.unitHash_ = murmur32.digest(unithash, 0)
        self.threadBuffer = threading.local()
        self.threadBuffer.value = bytearray(12)

    def assign(self, split: list, seed_hi: int, seed_lo: int):
        prob = self.probability(seed_hi, seed_lo)
        return self.choose_variant(split, prob)

    @staticmethod
    def choose_variant(split: list, prob: float):
        cum_sum = 0.0
        for index, item in enumerate(split):
            cum_sum += item
            if prob < cum_sum:
                return index

        return len(split) - 1

    def probability(self, seed_hi: int, seed_lo: int):
        buff = self.threadBuffer.value
        buffers.put_uint32(buff, 0, seed_lo)
        buffers.put_uint32(buff, 4, seed_hi)
        buffers.put_uint32(buff, 8, self.unitHash_)
        hashing = murmur32.digest(buff, 0)
        return (hashing & 0xffffffff) * VariantAssigner.__normalizer

    __normalizer = 1.0 / 0xffffffff



from typing import Optional

import jsons
from jsons import DeserializationError

from sdk.context import Context
from sdk.variable_parser import VariableParser


class DefaultVariableParser(VariableParser):

    def parse(self,
              context: Context,
              experiment_name: str,
              variant_name: str,
              config: str) -> Optional[dict]:
        try:
            return jsons.loads(config, dict)
        except DeserializationError:
            return None



import decimal
from decimal import Decimal

from sdk.jsonexpr.evaluator import Evaluator


def compare_to(this, that):
    if this > that:
        return 1
    elif this < that:
        return -1
    elif this == that:
        return 0


class ExprEvaluator(Evaluator):

    def __init__(self, operators: dict, vars: dict):
        self.vars = vars
        self.operators = operators

    def evaluate(self, expr: object):
        if type(expr) is list:
            return self.operators["and"].evaluate(self, expr)
        elif type(expr) is dict:
            for key, value in expr.items():
                if key not in self.operators:
                    return None
                op = self.operators[key]
                if op is not None:
                    res = op.evaluate(self, value)
                    return res
                break
        return None

    def boolean_convert(self, x: object):
        if type(x) is bool:
            return x
        elif type(x) is str:
            return x != "False" and x != "0" and x != ""
        elif type(x) is int or type(x) is float or type(x) is complex:
            return x != 0
        return x is not None

    def number_convert(self, x: object):
        if type(x) is int or type(x) is float or type(x) is complex:
            return x
        elif type(x) is bool:
            return 1.0 if x is True else 0.0
        elif type(x) is str:
            try:
                return Decimal(x)
            except decimal.InvalidOperation:
                return None
        return None

    def string_convert(self, x: object):
        if type(x) is str:
            return x
        elif type(x) is bool:
            return str(x)
        elif type(x) is int or type(x) is float or type(x) is complex:
            return str(x)
        return None

    def extract_var(self, path: str):
        frags = path.split("/")

        target = self.vars if self.vars is not None else {}
        for frag in frags:
            value = None
            if type(target) is list:
                try:
                    value = target[int(frag)]
                except BaseException as err:
                    print(err)
            elif type(target) is dict:
                if frag not in target:
                    return None
                value = target[frag]
            if value is not None:
                target = value
                continue
            return None

        return target

    def compare(self, lhs: object, rhs: object):
        if lhs is None:
            return 0 if rhs is None else None
        elif rhs is None:
            return None
        if type(lhs) is int or type(lhs) is float or type(lhs) is complex:
            rvalue = self.number_convert(rhs)
            if rvalue is not None:
                return compare_to(lhs, rvalue)
        elif type(lhs) is str:
            rvalue = self.string_convert(rhs)
            if rvalue is not None:
                return compare_to(lhs, rvalue)
        elif type(lhs) is bool:
            rvalue = self.boolean_convert(rhs)
            if rvalue is not None:
                return compare_to(lhs, rvalue)
        elif type(lhs) == type(rhs) and lhs == rhs:
            return 0
        return None



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.boolean_combinator import BooleanCombinator


class AndCombinator(BooleanCombinator):

    def combine(self, evaluator: Evaluator, exprs: list):
        for item in exprs:
            if not evaluator.boolean_convert(evaluator.evaluate(item)):
                return False
        return True



from sdk.internal.lock.read_write_lock import ReadWriteLock


class Concurrency:

    @staticmethod
    def add_rw(lock: ReadWriteLock, lst: list, value: object):
        try:
            lock.acquire_write()
            lst.append(value)
        finally:
            lock.release_write()

    @staticmethod
    def compute_if_absent_rw(lock: ReadWriteLock,
                             mp: dict,
                             key: object, computer):
        try:
            lock.acquire_read()
            if key in mp:
                return mp[key]
        finally:
            lock.release_read()

        try:
            lock.acquire_write()
            if key in mp:
                return mp[key]

            new_value = computer(key)
            mp[key] = new_value
            return new_value
        finally:
            lock.release_write()

    @staticmethod
    def get_rw(lock: ReadWriteLock, mp: dict, key: object):
        try:
            lock.acquire_write()
            if key not in mp:
                return None
            else:
                return mp[key]
        finally:
            lock.release_write()

    @staticmethod
    def put_rw(lock: ReadWriteLock, mp: dict, key: object, value: object):
        try:
            lock.acquire_write()
            previous = None
            if key in mp:
                previous = mp[key]
            mp[key] = value
            return previous
        finally:
            lock.release_write()



import sys as _sys

if _sys.version_info > (3, 0):
    def xrange(a, b, c):
        return range(a, b, c)

    def xencode(x):
        if isinstance(x, bytes) or isinstance(x, bytearray):
            return x
        else:
            return x.encode()
else:
    def xencode(x):
        return x
del _sys


def digest(key, seed):
    key = bytearray(xencode(key))

    length = len(key)
    nblocks = int(length / 4)

    h1 = seed

    c1 = 0xcc9e2d51
    c2 = 0x1b873593

    for block_start in xrange(0, nblocks * 4, 4):
        k1 = key[block_start + 3] << 24 | \
             key[block_start + 2] << 16 | \
             key[block_start + 1] << 8 | \
             key[block_start + 0]

        k1 = (c1 * k1) & 0xFFFFFFFF
        k1 = rotate_right(k1, 15)
        k1 = (c2 * k1) & 0xFFFFFFFF

        h1 ^= k1
        h1 = rotate_right(h1, 13)
        h1 = (h1 * 5 + 0xe6546b64) & 0xFFFFFFFF

    tail_index = nblocks * 4
    k1 = 0
    tail_size = length & 3

    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size > 0:
        k1 = (k1 * c1) & 0xFFFFFFFF
        k1 = rotate_right(k1, 15)
        k1 = (k1 * c2) & 0xFFFFFFFF
        h1 ^= k1

    unsigned_val = fmix(h1 ^ length)
    if unsigned_val & 0x80000000 == 0:
        return unsigned_val
    else:
        return -((unsigned_val ^ 0xFFFFFFFF) + 1)


def fmix(h: int):
    h ^= h >> 16
    h = (h * 0x85ebca6b) & 0xFFFFFFFF
    h ^= h >> 13
    h = (h * 0xc2b2ae35) & 0xFFFFFFFF
    h ^= h >> 16
    return h


def rotate_right(n, d):
    return (n << d) | (n >> (32 - d)) & 0xFFFFFFFF


def to_signed32(n):
    n = n & 0xffffffff
    return (n ^ 0x80000000) - 0x80000000



import re

from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.binary_operator import BinaryOperator


class MatchOperator(BinaryOperator):
    def binary(self, evaluator: Evaluator, lhs: object, rhs: object):
        text = evaluator.string_convert(lhs)
        if text is not None:
            pattern = evaluator.string_convert(rhs)
            if pattern is not None:
                compiled = re.compile(pattern)
                return bool(compiled.match(text))
        return None




def put_uint32(buf: bytearray, offset: int, x: int):
    buf[offset] = x & 0xff
    buf[offset + 1] = (x >> 8) & 0xff
    buf[offset + 2] = (x >> 16) & 0xff
    buf[offset + 3] = (x >> 24) & 0xff


def get_uint32(buf: bytearray, offset: int):
    return (buf[offset] & 0xff) \
           | ((buf[offset + 1] & 0xff) << 8) \
           | ((buf[offset + 2] & 0xff) << 16) \
           | ((buf[offset + 3] & 0xff) << 24)


def get_uint24(buf: bytearray, offset: int):
    return (buf[offset] & 0xff) \
           | ((buf[offset + 1] & 0xff) << 8) \
           | ((buf[offset + 2] & 0xff) << 16)


def get_uint16(buf: bytearray, offset: int):
    return (buf[offset] & 0xff) \
           | ((buf[offset + 1] & 0xff) << 8)


def get_uint8(buf: bytearray, offset: int):
    return buf[offset] & 0xff



from concurrent.futures import Future
from typing import Optional

from sdk.client import Client
from sdk.context_data_provider import ContextDataProvider
from sdk.json.context_data import ContextData


class DefaultContextDataProvider(ContextDataProvider):

    def __init__(self, client: Client):
        self.client = client

    def get_context_data(self) -> Future[Optional[ContextData]]:
        return self.client.get_context_data()



import os
import threading
import time
import typing
import unittest
from concurrent.futures import Future

from sdk.context_config import ContextConfig

from sdk.context import Context

from sdk.client import Client
from sdk.client_config import ClientConfig
from sdk.default_context_event_handler import DefaultContextEventHandler

from sdk.default_context_data_provider import DefaultContextDataProvider

from sdk.audience_matcher import AudienceMatcher
from sdk.default_http_client import DefaultHTTPClient
from sdk.default_http_client_config import DefaultHTTPClientConfig

from sdk.default_variable_parser import DefaultVariableParser

from sdk.context_event_logger import ContextEventLogger, EventType

from sdk.context_event_handler import ContextEventHandler

from sdk.context_data_provider import ContextDataProvider

from sdk.default_audience_deserializer import DefaultAudienceDeserializer

from sdk.default_context_data_deserializer import \
    DefaultContextDataDeserializer
from sdk.json.attribute import Attribute
from sdk.json.context_data import ContextData
from sdk.json.exposure import Exposure
from sdk.json.goal_achievement import GoalAchievement
from sdk.json.publish_event import PublishEvent
from sdk.json.unit import Unit
from sdk.time.clock import Clock
from sdk.time.fixed_clock import FixedClock


class ContextEventLoggerTest(ContextEventLogger):
    logger_data: typing.Optional[object] = None
    logger_type: typing.Optional[EventType] = None

    def handle_event(self, event_type: EventType, data: object):
        self.logger_data = data
        self.logger_type = event_type


class ClientContextMock(Client):
    def get_context_data(self):
        future = Future()
        context_data = ContextData()
        context_data.experiments = []
        future.set_result(context_data)
        return future

    def publish(self, event: PublishEvent):
        future = Future()
        future.set_result(None)
        return future


def string_in_list(st: str, lst: list):
    for val in lst:
        if val.name == st:
            return True
    return False


class ContextTest(unittest.TestCase):

    expectedVariants = {
        "exp_test_ab": 1,
        "exp_test_abc": 2,
        "exp_test_not_eligible": 0,
        "exp_test_fullon": 2,
        "exp_test_new": 1,
    }

    expectedVariables = {
        "banner.border": 1.0,
        "banner.size": "large",
        "button.color": "red",
        "submit.color": "blue",
        "submit.shape": "rect",
        "show-modal": True,
    }

    variableExperiments = {
        "banner.border": "exp_test_ab",
        "banner.size": "exp_test_ab",
        "button.color": "exp_test_abc",
        "card.width": "exp_test_not_eligible",
        "submit.color": "exp_test_fullon",
        "submit.shape": "exp_test_fullon",
        "show-modal": "exp_test_new",
    }

    units = {
        "session_id": "e791e240fcd3df7d238cfc285f475e8152fcc0ec",
        "user_id": "123456789",
        "email": "bleh@absmartly.com"
    }

    deser = DefaultContextDataDeserializer()
    audeser = DefaultAudienceDeserializer()
    data: ContextData
    refresh_data: ContextData
    audience_strict_data: ContextData
    data_future: Future
    data_future_strict: Future
    data_future_refresh: Future
    data_future_ready: Future
    data_future_failed: Future
    clock: Clock
    data_provider: ContextDataProvider
    event_handler: ContextEventHandler
    event_logger: ContextEventLogger = ContextEventLoggerTest()
    variable_parser: DefaultVariableParser
    audience_matcher: AudienceMatcher

    def set_up(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'res/context.json'),
                  'r') as file:
            content = file.read()
        with open(os.path.join(os.path.dirname(__file__),
                               'res/context-strict.json'),
                  'r') as file:
            content_strict = file.read()
        with open(os.path.join(os.path.dirname(__file__),
                               'res/refreshed.json'),
                  'r') as file:
            refreshed = file.read()
        self.data = self.deser.deserialize(
            bytes(content, encoding="utf-8"),
            0,
            len(content))
        self.audience_strict_data = self.deser.deserialize(
            bytes(content_strict, encoding="utf-8"),
            0,
            len(content_strict))
        self.refresh_data = self.deser.deserialize(
            bytes(refreshed, encoding="utf-8"),
            0,
            len(refreshed))
        self.data_future_ready = Future()
        self.data_future_ready.set_result(self.data)
        self.data_future = Future()
        self.data_future_failed = Future()
        self.data_future_failed.set_exception(RuntimeError("FAILED"))
        self.data_future_strict = Future()
        self.data_future_strict.set_result(self.audience_strict_data)
        self.data_future_refresh = Future()
        self.data_future_refresh.set_result(self.refresh_data)

        self.clock = FixedClock(1_620_000_000_000)
        client_config = ClientConfig()
        client_config.endpoint = "https://sandbox.test.io/v1"
        client_config.api_key = "gfsgsgsf"
        client_config.application = "www"
        client_config.environment = "test"
        default_client_config = DefaultHTTPClientConfig()
        default_client = DefaultHTTPClient(default_client_config)
        self.client = ClientContextMock(client_config, default_client)
        self.data_provider = DefaultContextDataProvider(self.client)
        self.event_handler = DefaultContextEventHandler(self.client)
        self.variable_parser = DefaultVariableParser()
        self.audience_matcher = AudienceMatcher(self.audeser)

    def create_test_context(self, config, data_future):
        return Context(self.clock,
                       config, data_future,
                       self.data_provider,
                       self.event_handler,
                       self.event_logger,
                       self.variable_parser,
                       self.audience_matcher)

    def test_constructor_sets_overrides(self):
        self.set_up()
        overrides = {"exp_test": 2, "exp_test_1": 1}
        config = ContextConfig()
        config.units = self.units
        config.overrides = overrides

        context = self.create_test_context(config, self.data_future)
        for key, value in overrides.items():
            res = context.get_override(key)
            self.assertEqual(value, res)

    def test_constructor_sets_custom_assignments(self):
        self.set_up()
        cassignments = {"exp_test": 2, "exp_test_1": 1}
        config = ContextConfig()
        config.units = self.units
        config.cassigmnents = cassignments

        context = self.create_test_context(config, self.data_future)
        for key, value in cassignments.items():
            res = context.get_custom_assignment(key)
            self.assertEqual(value, res)

    def test_becomes_ready_with_completed_future(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        context.close()

    def test_becomes_ready_exceptionally_with_completed_future(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_failed)
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data.experiments, None)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(True, context.is_failed())
        context.close()

    def test_becomes_ready_with_exception(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.data_future.set_exception(RuntimeError("FAILED"))
        context.wait_until_ready()
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data.experiments, None)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(True, context.is_failed())
        context.close()

    def test_becomes_ready_without_exception(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.data_future.set_result(self.data)
        context.wait_until_ready()
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        context.close()

    def test_wait_until_ready(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        def set_result():
            self.data_future.set_result(self.data)
        th = threading.Thread(target=set_result)
        th.start()

        context.wait_until_ready()
        data = context.get_data()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        context.close()

    def test_wait_until_ready_completed(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.wait_until_ready()
        data = context.get_data()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        context.close()

    def test_wait_until_ready_async(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())
        future = context.wait_until_ready_async()
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        def set_result():
            self.data_future.set_result(self.data)
        th = threading.Thread(target=set_result)
        th.start()

        future.result()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        context.close()

    def test_wait_until_ready_async_completed(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        future = context.wait_until_ready_async()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        future.result()
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)
        context.close()

    def test_error_when_closing(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)

        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)

        context.track("goal1", {"amount": 125, "hours": 245})
        self.assertEqual(False, context.is_closing())
        self.assertEqual(False, context.is_closed())

        def sl(event):
            future = Future()

            def set_result():
                time.sleep(0.2)
                future.set_result(None)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.publish = sl

        context.close_async()

        self.assertEqual(True, context.is_closing())
        self.assertEqual(False, context.is_closed())
        try:
            context.set_attribute("attr1", "value1")
        except RuntimeError as e:
            self.assertIsNotNone(e)
            self.assertEqual("ABSmartly Context is closing", str(e))
        time.sleep(0.3)
        context.close()

    def test_error_when_closed(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)

        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data, self.data)

        context.track("goal1", {"amount": 125, "hours": 245})
        self.assertEqual(False, context.is_closing())
        self.assertEqual(False, context.is_closed())

        def sl(event):
            future = Future()

            def set_result():
                time.sleep(0.2)
                future.set_result(None)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.publish = sl

        context.close()

        self.assertEqual(False, context.is_closing())
        self.assertEqual(True, context.is_closed())
        try:
            context.set_attribute("attr1", "value1")
        except RuntimeError as e:
            self.assertIsNotNone(e)
            self.assertEqual("ABSmartly Context is closed", str(e))
        time.sleep(0.3)
        context.close()

    def test_get_experiments(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        data = context.get_data()
        self.assertIsNotNone(data)
        self.assertEqual(data.experiments, self.data.experiments)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        context.close()

    def test_refresh_timer_when_ready(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.data_future.set_result(self.data)

        def sl():
            future = Future()

            def set_result():
                time.sleep(0.2)
                data = ContextData()
                data.experiments = []
                future.set_result(data)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.get_context_data = sl

        futu = context.refresh_async()
        self.assertEqual(True, context.refreshing.value)
        context.wait_until_ready()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        futu.result()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertEqual(False, context.refreshing.value)
        context.close()

    def test_unit_empty(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        try:
            context.set_unit("db_user_id", "")
        except ValueError as e:
            self.assertEqual("Unit UID must not be blank.", str(e))

        try:
            context.set_unit("session_id", "1")
        except ValueError as e:
            self.assertEqual("Unit already set.", str(e))

        context.close()

    def test_set_attributes(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_attribute("db_user_id", "test")
        context.set_attributes({"db_user_id2": "test2"})

        self.assertEqual(context.is_failed(), False)
        context.close()

    def test_set_overrides(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_override("db_user_id", 1)
        self.assertEqual(1, context.get_override("db_user_id"))
        context.set_overrides({"db_user_id2": 1})
        self.assertEqual(1, context.get_override("db_user_id2"))

        self.data_future.set_result(self.data)
        context.wait_until_ready()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertEqual(1, context.get_override("db_user_id"))
        self.assertEqual(1, context.get_override("db_user_id2"))

        context.close()

    def test_set_overrides_ready(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_override("db_user_id", 1)
        self.assertEqual(1, context.get_override("db_user_id"))
        context.set_overrides({"db_user_id2": 1})
        self.assertEqual(1, context.get_override("db_user_id2"))

        self.data_future.set_result(self.data)
        context.wait_until_ready()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertEqual(1, context.get_override("db_user_id"))
        self.assertEqual(1, context.get_override("db_user_id2"))

        self.assertIsNone(context.get_override("db_user_id3"))
        context.close()

    def test_assignment_does_not_override_full_on_or_not_assignments(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_custom_assignment("exp_test_not_eligible", 3)
        context.set_custom_assignment("exp_test_fullon", 3)

        res = context.get_treatment("exp_test_not_eligible")
        self.assertEqual(0, res)
        res = context.get_treatment("exp_test_fullon")
        self.assertEqual(2, res)
        self.assertEqual(2, context.get_pending_count())
        self.assertEqual(3, context.get_custom_assignment("exp_test_fullon"))

        self.assertEqual(
            3,
            context.get_custom_assignment("exp_test_not_eligible"))
        self.assertIsNone(context.get_custom_assignment("db_user_id3"))
        context.close()

    def test_set_overrides_clear_assignment_cache(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_override("db_user_id", 1)
        self.assertEqual(1, context.get_override("db_user_id"))
        context.set_overrides({"db_user_id2": 1})
        self.assertEqual(1, context.get_override("db_user_id2"))

        self.data_future.set_result(self.data)
        context.wait_until_ready()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertEqual(1, context.get_override("db_user_id"))
        self.assertEqual(1, context.get_override("db_user_id2"))

        res = context.get_treatment("db_user_id")
        self.assertEqual(res, context.get_override("db_user_id"))
        self.assertEqual(1, context.get_pending_count())

        self.assertIsNone(context.get_override("db_user_id3"))
        context.close()

    def test_set_custom_assignments_ready(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_custom_assignment("db_user_id", 1)
        self.assertEqual(1, context.get_custom_assignment("db_user_id"))
        context.set_custom_assignments({"db_user_id2": 1})
        self.assertEqual(1, context.get_custom_assignment("db_user_id2"))

        self.assertIsNone(context.get_custom_assignment("db_user_id3"))
        context.close()

    def test_set_custom_assignments(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_custom_assignment("db_user_id", 1)
        self.assertEqual(1, context.get_custom_assignment("db_user_id"))
        context.set_custom_assignments({"db_user_id2": 1})
        self.assertEqual(1, context.get_custom_assignment("db_user_id2"))

        self.data_future.set_result(self.data)
        context.wait_until_ready()
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        self.assertEqual(1, context.get_custom_assignment("db_user_id"))
        self.assertEqual(1, context.get_custom_assignment("db_user_id2"))

        self.assertIsNone(context.get_custom_assignment("db_user_id3"))
        context.close()

    def test_set_custom_assignments_pending_assignment_cache(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_custom_assignment("exp_test_ab", 2)
        context.set_custom_assignments({"exp_test_abc": 3})

        self.assertEqual(0, context.get_pending_count())
        res = context.get_treatment("exp_test_ab")
        self.assertEqual(2, res)
        self.assertEqual(1, context.get_pending_count())

        res = context.get_treatment("exp_test_abc")
        self.assertEqual(3, res)
        self.assertEqual(2, context.get_pending_count())

        context.set_custom_assignment("exp_test_ab", 4)
        res = context.get_treatment("exp_test_ab")
        self.assertEqual(4, res)
        self.assertEqual(3, context.get_pending_count())

        context.close()

    def test_peek_treatment(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        for experiment in self.data.experiments:
            res = context.peek_treatment(experiment.name)
            self.assertEqual(self.expectedVariants[experiment.name], res)

        res = context.peek_treatment("not_found")
        self.assertEqual(0, res)
        context.close()

    def test_peek_variable(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        for key, value in self.variableExperiments.items():
            res = context.peek_variable_value(key, 17)
            if value != "exp_test_not_eligible" and \
                    string_in_list(value, self.data.experiments):
                self.assertEqual(self.expectedVariables[key], res)
            else:
                self.assertEqual(17, res)

        self.assertEqual(0, context.get_pending_count())
        context.close()

    def test_peek_variable_strict(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_strict)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        res = context.peek_variable_value("banner.size", "small")
        self.assertEqual("small", res)

        self.assertEqual(0, context.get_pending_count())
        context.close()

    def test_get_variable(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        for key, value in self.variableExperiments.items():
            res = context.get_variable_value(key, 17)
            if value != "exp_test_not_eligible" and \
                    string_in_list(value, self.data.experiments):
                self.assertEqual(self.expectedVariables[key], res)
            else:
                self.assertEqual(17, res)

        self.assertEqual(4, context.get_pending_count())
        context.close()

    def test_get_variable_strict(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_strict)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        res = context.get_variable_value("banner.size", "small")
        self.assertEqual("small", res)

        self.assertEqual(1, context.get_pending_count())
        context.close()

    def test_get_variable_keys(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_refresh)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        res = context.get_variable_keys()
        self.assertEqual(self.variableExperiments, res)

        self.assertEqual(0, context.get_pending_count())
        context.close()

    def test_get_custom_field_keys(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        res = context.get_custom_field_keys()
        self.assertEqual(["country", "languages", "overrides"], res)
        context.close()

    def test_get_custom_field_values(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        self.assertEqual(None, context.get_custom_field_value("not_found", "not_found"))
        self.assertEqual(None, context.get_custom_field_value("exp_test_ab", "not_found"))

        self.assertEqual("US,PT,ES,DE,FR", context.get_custom_field_value("exp_test_ab", "country"))
        self.assertEqual({'123': 1, '456': 0}, context.get_custom_field_value("exp_test_ab", "overrides"))
        self.assertEqual("json", context.get_custom_field_type("exp_test_ab", "overrides"))

        self.assertEqual(None, context.get_custom_field_value("exp_test_ab", "languages"))
        self.assertEqual(None, context.get_custom_field_type("exp_test_ab", "languages"))

        self.assertEqual(None, context.get_custom_field_value("exp_test_abc", "overrides"))
        self.assertEqual(None, context.get_custom_field_type("exp_test_abc", "overrides"))

        self.assertEqual("en-US,en-GB,pt-PT,pt-BR,es-ES,es-MX", context.get_custom_field_value("exp_test_abc", "languages"))
        self.assertEqual("string", context.get_custom_field_type("exp_test_abc", "languages"))

        self.assertEqual(None, context.get_custom_field_value("exp_test_no_custom_fields", "country"))
        self.assertEqual(None, context.get_custom_field_type("exp_test_no_custom_fields", "country"))

        self.assertEqual(None, context.get_custom_field_value("exp_test_no_custom_fields", "overrides"))
        self.assertEqual(None, context.get_custom_field_type("exp_test_no_custom_fields", "overrides"))

        self.assertEqual(None, context.get_custom_field_value("exp_test_no_custom_fields", "languages"))
        self.assertEqual(None, context.get_custom_field_type("exp_test_no_custom_fields", "languages"))

        context.close()

    def test_track(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.track("goal1", {"amount": 125, "hours": 245})
        self.assertEqual(1, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.publish()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.close()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(True, context.is_closed())

    def test_track_not_ready(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.track("goal1", {"amount": 125, "hours": 245})
        self.assertEqual(1, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.track("goal2", {"amount": 125, "hours": 245})
        self.assertEqual(2, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.publish()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.close()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(True, context.is_closed())

    def test_publish_resets_queue_and_keeps_attr_over_and_assignments(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.set_custom_assignment("exp_test_ab", 2)
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(2, context.get_treatment("exp_test_ab"))

        context.track("goal2", {"amount": 125, "hours": 245})
        self.assertEqual(2, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.publish()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_closed())

        self.assertEqual(2, context.get_custom_assignment("exp_test_ab"))
        context.close()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(True, context.is_closed())

    def test_starts_publish_timeout_when_ready_with_queue_not_empty(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future)
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_failed())

        context.track("goal2", {"amount": 125, "hours": 245})
        self.assertEqual(1, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_closed())

        context.publish()
        self.assertEqual(None, context.timeout)
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(False, context.is_ready())
        self.assertEqual(False, context.is_closed())
        self.data_future.set_result(self.data)
        context.wait_until_ready()
        self.assertEqual(None, context.get_custom_assignment("exp_test_ab"))
        context.close()
        self.assertEqual(0, context.get_pending_count())
        self.assertEqual(True, context.is_ready())
        self.assertEqual(True, context.is_closed())

    def test_publish_success(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        context.set_attribute("test", "value1")
        context.set_custom_assignment("testAssignment", 5)
        res = context.get_treatment("testAssignment")
        self.assertEqual(0, res)

        event = PublishEvent()
        event.hashed = True
        event.units = []
        unit = Unit()
        unit.type = "session_id"
        unit.uid = "pAE3a1i5Drs5mKRNq56adA"
        unit1 = Unit()
        unit1.type = "user_id"
        unit1.uid = "JfnnlDI7RTiF9RgfG2JNCw"
        unit2 = Unit()
        unit2.type = "email"
        unit2.uid = "IuqYkNRfEx5yClel4j3NbA"
        event.units.append(unit)
        event.units.append(unit1)
        event.units.append(unit2)
        event.publishedAt = 1620000000000
        exposure = Exposure()
        exposure.id = 0
        exposure.name = "testAssignment"
        exposure.unit = ""
        exposure.variant = 0
        exposure.exposedAt = 1620000000000
        exposure.eligible = True
        event.exposures = []
        event.exposures.append(exposure)
        goal = GoalAchievement()
        goal.name = "goal1"
        goal.achievedAt = 1620000000000
        goal.properties = {"amount": 125}
        event.goals = []
        event.goals.append(goal)
        event.attributes = []
        attribute = Attribute()
        attribute.name = "test"
        attribute.value = "value1"
        attribute.setAt = 1620000000000
        event.attributes.append(attribute)

        def pubka(pub):
            self.assertEqual(event.goals[0].properties,
                             pub.goals[0].properties)
            self.assertEqual(event.goals[0].name,
                             pub.goals[0].name)
            self.assertEqual(event.attributes[0].name,
                             pub.attributes[0].name)
            self.assertEqual(event.attributes[0].value,
                             pub.attributes[0].value)
            self.assertEqual(event.exposures[0].name,
                             pub.exposures[0].name)
            self.assertEqual(event.exposures[0].variant,
                             pub.exposures[0].variant)
            self.assertEqual(event.units[0].uid,
                             pub.units[0].uid)
            self.assertEqual(len(event.exposures),
                             len(pub.exposures))
            self.assertEqual(len(event.goals),
                             len(pub.goals))
            self.assertEqual(len(event.units),
                             len(pub.units))
            self.assertEqual(len(event.attributes),
                             len(pub.attributes))
            future = Future()
            future.set_result(None)
            return future

        self.client.publish = pubka

        context.track("goal1", {"amount": 125})
        context.publish()
        context.close()

    def test_flush_mapper(self):
        result = "pAE3a1i5Drs5mKRNq56adA"
        force = result.encode('ascii', errors='ignore')\
            .decode()
        self.assertEqual(result, force)

        result = "pAE3a1i5Drs5%KRNq56adA"
        force = "pAE3a1i5Drs5%паKRNq56adA"\
            .encode('ascii', errors='ignore')\
            .decode()
        self.assertEqual(result, force)

    def test_refresh_async(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_refresh)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())

        def sl():
            future = Future()

            def set_result():
                future.set_result(self.data)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.get_context_data = sl
        self.assertTrue(context.refresh_timer is not None)
        fut = context.refresh_async()
        fut.result()
        self.assertTrue(context.refresh_timer is not None)
        self.assertEqual(True, context.is_ready())
        self.assertEqual(False, context.is_failed())
        res = context.get_experiments()
        self.assertEqual("exp_test_ab", res[0])
        context.close()

    def test_refresh_clear_assignment_cache_for_started_experiment(self):
        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(True, context.is_ready())

        res = context.get_treatment("exp_test_new")
        self.assertEqual(0, res)
        self.assertEqual(1, context.get_pending_count())

        def sl():
            future = Future()

            def set_result():
                future.set_result(self.data)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.get_context_data = sl

        context.refresh()
        res = context.get_experiments()
        self.assertEqual("exp_test_ab", res[0])
        self.assertEqual("exp_test_fullon", res[3])
        self.assertEqual(4, len(res))
        res = context.get_treatment("exp_test_fullon")
        self.assertEqual(2, res)
        self.assertEqual(2, context.get_pending_count())
        context.close()

    def test_logger_called(self):
        self.event_logger.logger_data = None
        self.event_logger.logger_type = None
        self.assertIsNone(self.event_logger.logger_data)
        self.assertIsNone(self.event_logger.logger_type)

        self.set_up()
        config = ContextConfig()
        config.units = self.units
        context = self.create_test_context(config, self.data_future_ready)
        self.assertEqual(self.event_logger.logger_type, EventType.READY)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(ContextData()))
        self.assertEqual(True, context.is_ready())

        res = context.get_treatment("exp_test_new")
        self.assertEqual(0, res)
        self.assertEqual(1, context.get_pending_count())
        self.assertEqual(self.event_logger.logger_type, EventType.EXPOSURE)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(Exposure()))

        def sl():
            future = Future()

            def set_result():
                future.set_result(self.data)
            th = threading.Thread(target=set_result)
            th.start()
            return future
        self.client.get_context_data = sl
        self.assertEqual(self.event_logger.logger_type, EventType.EXPOSURE)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(Exposure()))

        context.refresh()
        self.assertEqual(self.event_logger.logger_type, EventType.REFRESH)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(ContextData()))
        res = context.get_experiments()
        self.assertEqual(self.event_logger.logger_type, EventType.REFRESH)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(ContextData()))
        self.assertEqual("exp_test_ab", res[0])
        self.assertEqual("exp_test_fullon", res[3])
        self.assertEqual(4, len(res))
        res = context.get_treatment("exp_test_fullon")
        self.assertEqual(self.event_logger.logger_type, EventType.EXPOSURE)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(Exposure()))
        self.assertEqual(2, res)
        self.assertEqual(2, context.get_pending_count())
        self.assertEqual(self.event_logger.logger_type, EventType.EXPOSURE)
        self.assertEqual(type(self.event_logger.logger_data),
                         type(Exposure()))
        context.close()
        self.assertEqual(self.event_logger.logger_type, EventType.CLOSE)
        self.assertIsNone(self.event_logger.logger_data)



import threading


class ReadWriteLock:

    def __init__(self):
        self.w_lock = threading.RLock()
        self.num_r_lock = threading.RLock()
        self.num_r = 0

    def acquire_read(self):
        self.num_r_lock.acquire()
        self.num_r += 1
        if self.num_r == 1:
            self.w_lock.acquire()
        self.num_r_lock.release()

    def release_read(self):
        assert self.num_r > 0
        self.num_r_lock.acquire()
        self.num_r -= 1
        if self.num_r == 0:
            self.w_lock.release()
        self.num_r_lock.release()

    def acquire_write(self):
        self.w_lock.acquire()

    def release_write(self):
        self.w_lock.release()



class DefaultHTTPClientConfig:
    def __init__(self):
        self.connection_timeout = 3  # seconds
        self.connection_request_timeout = 1
        self.retry_interval = 0.3
        self.max_retries = 5



import os
import unittest

from sdk.default_context_data_deserializer import \
    DefaultContextDataDeserializer
from sdk.json.context_data import ContextData
from sdk.json.experiement_application import ExperimentApplication
from sdk.json.experiment import Experiment
from sdk.json.experiment_variant import ExperimentVariant


class ContextDataDeserializerTest(unittest.TestCase):

    def test_deserialize(self):
        with open(os.path.join(
                os.path.dirname(__file__),
                'res/context.json'),
                'r') as file:
            data = file.read()
        deser = DefaultContextDataDeserializer()
        result = deser.deserialize(bytes(data, encoding="utf-8"), 0, len(data))
        exp_appl = ExperimentApplication()
        exp_appl.name = "website"

        varone = ExperimentVariant()
        varone.name = "A"
        varone.config = None
        vartwo = ExperimentVariant()
        vartwo.name = "0"
        vartwo.config = "{\"banner.border\":1,\"banner.size\":\"large\"}"

        experiment0 = Experiment()
        experiment0.id = 1
        experiment0.name = "exp_test_ab"
        experiment0.unitType = "session_id"
        experiment0.iteration = 1
        experiment0.seedHi = 3603515
        experiment0.seedLo = 233373850
        experiment0.split = [0.5, 0.5]
        experiment0.trafficSeedHi = 449867249
        experiment0.trafficSeedLo = 455443629
        experiment0.trafficSplit = [0.0, 1.0]
        experiment0.fullOnVariant = 0
        experiment0.applications = [exp_appl]
        experiment0.variants = [
            varone,
            vartwo
        ]
        experiment0.audienceStrict = False
        experiment0.audience = None

        vartwot = ExperimentVariant()
        vartwot.name = "0"
        vartwot.config = "{\"button.color\":\"blue\"}"

        vartwott = ExperimentVariant()
        vartwott.name = "C"
        vartwott.config = "{\"button.color\":\"red\"}"

        experiment1 = Experiment()
        experiment1.id = 2
        experiment1.name = "exp_test_abc"
        experiment1.unitType = "session_id"
        experiment1.iteration = 1
        experiment1.seedHi = 55006150
        experiment1.seedLo = 47189152
        experiment1.split = [0.34, 0.33, 0.33]
        experiment1.trafficSeedHi = 705671872
        experiment1.trafficSeedLo = 212903484
        experiment1.trafficSplit = [0.0, 1.0]
        experiment1.fullOnVariant = 0
        experiment1.applications = [exp_appl]
        experiment1.variants = [
            varone,
            vartwot,
            vartwott
        ]
        experiment1.audienceStrict = False
        experiment1.audience = ""

        vartwotta = ExperimentVariant()
        vartwotta.name = "B"
        vartwotta.config = "{\"card.width\":\"80%\"}"
        vartwottaa = ExperimentVariant()
        vartwottaa.name = "C"
        vartwottaa.config = "{\"card.width\":\"75%\"}"

        experiment2 = Experiment()
        experiment2.id = 3
        experiment2.name = "exp_test_not_eligible"
        experiment2.unitType = "user_id"
        experiment2.iteration = 1
        experiment2.seedHi = 503266407
        experiment2.seedLo = 144942754
        experiment2.split = [0.34, 0.33, 0.33]
        experiment2.trafficSeedHi = 87768905
        experiment2.trafficSeedLo = 511357582
        experiment2.trafficSplit = [0.99, 0.01]
        experiment2.fullOnVariant = 0
        experiment2.applications = [exp_appl]
        experiment2.variants = [
            varone,
            vartwotta,
            vartwottaa
        ]
        experiment2.audienceStrict = False
        experiment2.audience = "{}"

        vartwottab = ExperimentVariant()
        vartwottab.name = "B"
        vartwottab.config = \
            "{\"submit.color\":\"red\",\"submit.shape\":\"circle\"}"

        vartwottabc = ExperimentVariant()
        vartwottabc.name = "C"
        vartwottabc.config = \
            "{\"submit.color\":\"blue\",\"submit.shape\":\"rect\"}"

        vartwottabb = ExperimentVariant()
        vartwottabb.name = "D"
        vartwottabb.config = \
            "{\"submit.color\":\"green\",\"submit.shape\":\"square\"}"

        experiment3 = Experiment()
        experiment3.id = 4
        experiment3.name = "exp_test_fullon"
        experiment3.unitType = "session_id"
        experiment3.iteration = 1
        experiment3.seedHi = 856061641
        experiment3.seedLo = 990838475
        experiment3.split = [0.25, 0.25, 0.25, 0.25]
        experiment3.seedHi = 856061641
        experiment3.seedLo = 990838475
        experiment3.trafficSplit = [0.0, 1.0]
        experiment3.fullOnVariant = 2
        experiment3.applications = [exp_appl]
        experiment3.variants = [
            varone,
            vartwottab,
            vartwottabc,
            vartwottabb
        ]
        experiment3.audienceStrict = False
        experiment3.audience = "null"

        expected = ContextData()
        expected.experiments = [experiment0,
                                experiment1,
                                experiment2,
                                experiment3]

        self.assertIsNotNone(result)
        self.assertEqual(len(result.experiments),
                         len(expected.experiments))
        self.assertEqual(result.experiments[3].variants[3].config,
                         expected.experiments[3].variants[3].config)



from abc import abstractmethod
from concurrent.futures import Future
from typing import Optional

from sdk.json.context_data import ContextData


class ContextDataProvider:

    @abstractmethod
    def get_context_data(self) -> Future[Optional[ContextData]]:
        raise NotImplementedError



from abc import abstractmethod

from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operator import Operator


class UnaryOperator(Operator):
    def evaluate(self, evaluator: Evaluator, args: object):
        arg = evaluator.evaluate(args)
        return self.unary(evaluator, arg)

    @abstractmethod
    def unary(self, evaluator: Evaluator, arg: object):
        raise NotImplementedError



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.binary_operator import BinaryOperator


class InOperator(BinaryOperator):
    def binary(self, evaluator: Evaluator, haystack: object, needle: object):
        if type(haystack) is list:
            for item in haystack:
                if evaluator.compare(item, needle) == 0:
                    return True
            return False
        elif type(haystack) is str or type(haystack) is dict:
            needle_str = evaluator.string_convert(needle)
            return needle_str is not None and needle_str in haystack
        else:
            return None



import jsons

from sdk.context_event_serializer import ContextEventSerializer
from sdk.json.publish_event import PublishEvent


class DefaultContextEventSerializer(ContextEventSerializer):

    def serialize(self, publish_event: PublishEvent) -> bytearray:
        str_result = jsons.dumps(publish_event, strip_nulls=True)
        return bytearray(str_result, encoding='utf-8')



import os
from concurrent.futures import ThreadPoolExecutor

from sdk.default_context_data_deserializer import \
    DefaultContextDataDeserializer
from sdk.default_context_event_serializer import \
    DefaultContextEventSerializer


class ClientConfig:
    def __init__(self, prefix=""):
        self.executor = ThreadPoolExecutor()
        self.endpoint = os.environ.get(prefix + "endpoint")
        self.environment = os.environ.get(prefix + "environment")
        self.application = os.environ.get(prefix + "application")
        self.api_key = os.environ.get(prefix + "api_key")
        self.serializer = DefaultContextEventSerializer()
        self.deserializer = DefaultContextDataDeserializer()



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operators.binary_operator import BinaryOperator


class GreaterThanOrEqualOperator(BinaryOperator):
    def binary(self, evaluator: Evaluator, lhs: object, rhs: object):
        result = evaluator.compare(lhs, rhs)
        return result >= 0 if result is not None else None



from sdk.jsonexpr.evaluator import Evaluator
from sdk.jsonexpr.operator import Operator


class VarOperator(Operator):
    def evaluate(self, evaluator: Evaluator, path: object):
        if type(path) is dict:
            if "path" not in path:
                path = None
            else:
                path = path["path"]
        return evaluator.extract_var(path) if type(path) is str else None



from sdk.client_config import ClientConfig
from sdk.http_client import HTTPClient
from sdk.json.publish_event import PublishEvent


class Client:

    def __init__(self, config: ClientConfig, http_client: HTTPClient):
        self.serializer = config.serializer
        self.deserializer = config.deserializer
        self.executor = config.executor
        endpoint = config.endpoint
        api_key = config.api_key
        application = config.application
        environment = config.environment

        self.url = endpoint + "/context"
        self.http_client = http_client

        self.headers = {"X-API-Key": api_key,
                        "X-Application": application,
                        "X-Environment": environment,
                        "X-Application-Version": '0',
                        "X-Agent": "absmartly-python-sdk"}
        self.query = {"application": application,
                      "environment": environment}

    def get_context_data(self):
        return self.executor.submit(self.send_get, self.url, self.query, {})

    def send_get(self, url: str, query: dict, headers: dict):
        response = self.http_client.get(url, query, headers)
        if response.status_code // 100 == 2:
            content = response.content
            return self.deserializer.deserialize(content, 0, len(content))
        return response.raise_for_status()

    def publish(self, event: PublishEvent):
        return self.executor.submit(
            self.send_put,
            self.url,
            {},
            self.headers,
            event)

    def send_put(self,
                 url: str,
                 query: dict,
                 headers: dict,
                 event: PublishEvent):
        content = self.serializer.serialize(event)
        response = self.http_client.put(url, query, headers, content)
        if response.status_code // 100 == 2:
            content = response.content
            return self.deserializer.deserialize(content, 0, len(content))
        return response.raise_for_status()



from abc import abstractmethod
from typing import Optional


class VariableParser:

    @abstractmethod
    def parse(self,
              context,
              experiment_name: str,
              variant_name: str,
              variable_value: str) -> Optional[dict]:
        raise NotImplementedError
